/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Line state detection application for all channel types.
 * 
 * Copyright (C) 2004-2005, Newman Telecom, Inc.
 * Copyright (C) 2004-2005, Newman Ventures, Inc.
 * Copyright (C) 2005, NGM Tec, Inc.
 * 
 * Justin Newman <jnewman@newmantelecom.com>
 * 
 * Portions Copyright:
 * Copyright (C) 2001, Linux Support Services, Inc.
 * Copyright (C) 2004, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * Financial Contributions:
 * NGM Tec, Inc.
 * Newman Ventures, Inc.
 * 
 * We would like to thank NGM Tec, Inc. for funding this project
 * and allowing us to release it GPL.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */
 
#include <asterisk/lock.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/translate.h>
#include <asterisk/dsp.h>
#include <asterisk/utils.h>
#include <string.h>
#include <stdlib.h>

static char *tdesc = "Newman's line state detection application";

static char *app = "NVLineDetect";

static char *synopsis = "Detects answer/dead/other signals on ZAP and other channels";

static char *descrip = 
"  NVLineDetect([waitdur[|options[|ringdur[|deaddur[|sildur[|mindur[|maxdur]]]]]]]):\n"
"This application listens for certain tones (on ZAP and most channel types)\n"
"for max waitdur seconds of time. In particular, it detects presence of\n"
"conditions: ANSWER, PICKUP, RING, BUSY, CONGESTION, DIALTONE, and DEAD.\n"
"The respective extension name (in lowercase letters) will be called\n"
"(i.e. 'answer', 'pickup', etc.). In addition, the variable TONE_DETECTED\n"
"is set. If all undetected, control will continue at the next priority.\n"
"      waitdur:  Maximum number of seconds to wait (default=30)\n"
"      options:\n"
"        'n':  Attempt on-hook if unanswered (default=no)\n"
"        'd':  Detect answer detection of ring+talk (default=no)\n"
"        'a':  Ignore pickup detection of ring+sil (default=no)\n"
"        'b':  Ignore busy detection (default=no)\n"
"        'c':  Ignore congestion detection (default=no)\n"
"        'r':  Return after ring/ringing detection (default=no)\n"
"        'd':  Return after dialtone detection (default=no)\n"
"        's':  Ignore dead channel detection (default=no)\n"
"      ringdur:  Maximum ms of no activity between rings (default=4800)\n"
"      deaddur:  How many ms of no activity to wait for dead (default=10000)\n"
"      sildur:  Silence ms after min/maxdur before answer/pickup (default=100)\n"
"      mindur:  Minimum non-silence talk ms needed (default=20)\n"
"      maxdur:  Maximum non-silence talk ms allowed (default=0/forever)\n"
"Returns -1 on hangup, and 0 on successful completion with no exit conditions.\n\n"
"For questions or comments, please e-mail support@newmantelecom.com.\n";

#define DEBUG_MODE	LOG_DEBUG
#define CALLERID_FIELD cid.cid_num

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

int transfer_exten(struct ast_channel *chan, char *tone_type, char *extension)
{
	/* In case we continue, allow them to look at this */
	pbx_builtin_setvar_helper(chan, "TONE_DETECTED", tone_type);

	/* Now redirect off to the correct extension, or continue */
	ast_log(LOG_NOTICE, "Redirecting %s to %s extension\n", chan->name, extension);
	if (ast_exists_extension(chan, chan->context, extension, 1, chan->CALLERID_FIELD)) {
		strncpy(chan->exten, extension, sizeof(chan->exten) - 1);
		chan->priority = 0;
	} else {
		ast_log(LOG_WARNING, "Tone %s detected, but no %s extension\n", tone_type, extension);
	}	
	return 0;
}

static int nv_detectline_exec(struct ast_channel *chan, void *data)
{
	int res = 0;
	struct localuser *u;
	char tmp[256] = "\0";
	char *p = NULL;
	char *waitstr = NULL;
	char *options = NULL;
	char *ringstr = NULL;
	char *deadstr = NULL;
	char *silstr = NULL;
	char *minstr = NULL;
	char *maxstr = NULL;
	struct ast_frame *fr = NULL;
	struct ast_frame *fr2 = NULL;
	int notsilent = 0;
	struct timeval start = {0, 0}, end = {0, 0};
	int waitdur = 60;
	int ringdur = 4800;
	int deaddur = 10000;
	int sildur = 100;
	int mindur = 20;
	int maxdur = -1;
	int skipanswer = 0;
	int ignoreanswerdetect = 1;
	int ignorepickupdetect = 0;
	int ignorebusydetect = 0;
	int ignorecongestiondetect = 0;
	int returnonring = 0;
	int returnondialtone = 0;
	int ignoredeaddetect = 0;
	int gottalk = 0;
	int gotanswer = 0;
	int gotpickup = 0;
	int gotbusy = 0;
	int gotcongestion = 0;
	int gotring = 0;
	int gotdialtone = 0;
	int gotdead = 0;
	int x = 0;
	int origrformat = 0;
	long features = 0;
	time_t timeout = 0;
	struct ast_dsp *dsp = NULL;
    
	pbx_builtin_setvar_helper(chan, "TONE_DETECTED", "");

	if (data || !ast_strlen_zero((char *)data)) {
		strncpy(tmp, (char *)data, sizeof(tmp)-1);
	}   

	p = tmp;

	waitstr = strsep(&p, "|");
	options = strsep(&p, "|");
	ringstr = strsep(&p, "|");
	deadstr = strsep(&p, "|");
	silstr = strsep(&p, "|");
	minstr = strsep(&p, "|");
	maxstr = strsep(&p, "|");

	if (waitstr) {
		if ((sscanf(waitstr, "%d", &x) == 1) && (x > 0))
		waitdur = x;
	}

	if (ringstr) {
		if ((sscanf(ringstr, "%d", &x) == 1) && (x > 0))
		ringdur = x;
	}

	if (deadstr) {
		if ((sscanf(deadstr, "%d", &x) == 1) && (x > 0))
		deaddur = x;
	}

	if (silstr) {
		if ((sscanf(silstr, "%d", &x) == 1) && (x > 0))
		sildur = x;
	}

	if (minstr) {
		if ((sscanf(minstr, "%d", &x) == 1) && (x > 0))
		mindur = x;
	}

	if (maxstr) {
		if ((sscanf(maxstr, "%d", &x) == 1) && (x > 0))
		maxdur = x;
	}

	if (options) {
		if (strchr(options, 'n')) {
			skipanswer = 1;
		}
		if (strchr(options, 'd')) {
			ignoreanswerdetect = 0;
		}
		if (strchr(options, 'a')) {
			ignorepickupdetect = 1;
		}
		if (strchr(options, 'b')) {
			ignorebusydetect = 1;
		}
		if (strchr(options, 'c')) {
			ignorecongestiondetect = 1;
		}
		if (strchr(options, 'r')) {
			returnonring = 1;
		}
		if (strchr(options, 'd')) {
			returnondialtone = 1;
		}
		if (strchr(options, 's')) {
			ignoredeaddetect = 1;
		}
	}
    
#ifndef DSP_PROGRESS_PICKUP
	ast_log(LOG_WARNING, "PICKUP detection is not available without dsp.c/dsp.h/frame.h changes\n");
	ignorepickupdetect = 1;
#endif

#ifndef DSP_PROGRESS_DEAD
	ast_log(LOG_WARNING, "DEAD detection is not available without dsp.c/dsp.h/frame.h changes\n");
	ignoredeaddetect = 1;
#endif

#ifndef DSP_PROGRESS_DIALTONE
	ast_log(LOG_WARNING, "DIALTONE detection is not available without dsp.c/dsp.h/frame.h changes\n");
	returnondialtone = 0;
#endif
    

// TODO: handle stutter

	ast_log(DEBUG_MODE, "Preparing line detect '%s' (waitdur=%dsec, deadur=%dms, sildur=%dms, mindur=%dms, maxdur=%dms)\n", 
		tmp, waitdur, deaddur, sildur, mindur, maxdur);

	LOCAL_USER_ADD(u);
	if (chan->_state != AST_STATE_UP && !skipanswer) {
		/* Otherwise answer unless we're supposed to operate while on-hook */
		res = ast_answer(chan);
	}
	if (!res) {
		origrformat = chan->readformat;
		if ((res = ast_set_read_format(chan, AST_FORMAT_SLINEAR))) 
		ast_log(LOG_WARNING, "Unable to set read format to linear!\n");
	}
	if (!(dsp = ast_dsp_new())) {
		ast_log(LOG_WARNING, "Unable to allocate DSP!\n");
		res = -1;
	}
    
	if (dsp) {  
		features |= DSP_PROGRESS_BUSY;
		features |= DSP_PROGRESS_RINGING;
		features |= DSP_PROGRESS_CONGESTION;
	#ifdef DSP_PROGRESS_DIALTONE
		features |= DSP_PROGRESS_DIALTONE;
	#endif
	#ifdef DSP_PROGRESS_DEAD
		features |= DSP_PROGRESS_DEAD;
	#endif
	#ifdef DSP_PROGRESS_PICKUP
		features |= DSP_PROGRESS_PICKUP;
	#endif
		ast_dsp_set_threshold(dsp, 256); 
		ast_dsp_set_features(dsp, features | DSP_DIGITMODE_RELAXDTMF);
		ast_dsp_digitmode(dsp, DSP_DIGITMODE_DTMF);
		ast_dsp_set_deadthreshold(dsp, deaddur);
		ast_dsp_set_ringthreshold(dsp, ringdur);
		ast_dsp_reset_deaddetect(dsp);
		ast_dsp_reset_pickupdetect(dsp);
	}

	if (!res) {
		if (waitdur > 0) {
			timeout = time(NULL) + (time_t)waitdur;
		}

		if (deaddur >= waitdur*1000) {
			ast_log(LOG_WARNING, "Deaddur is equal or greater than waitdur (how can we test?)\n");
		}

		if (ringdur >= waitdur*1000) {
			ast_log(LOG_WARNING, "Ringdur is equal or greater than waitdur (how can we test?)\n");
		}

		while(ast_waitfor(chan, -1) > -1) {
			/* Did we flat out right time out? */
			if (waitdur > 0 && time(NULL) > timeout) {
				res = 0;
				break;
			}

			fr = ast_read(chan);
			if (!fr) {
				ast_log(DEBUG_MODE, "Received unexpected hangup\n");
				res = -1;
				break;
			}

			fr2 = ast_dsp_process(chan, dsp, fr);
			if (!fr2) {
				ast_log(LOG_WARNING, "DSP received is NULL (what happened?)\n");
				/* Continue, skipping the DSP message this round */
				fr2 = fr;
			} 

			if (fr2->frametype == AST_FRAME_CONTROL) {          
				if (fr2->subclass == AST_CONTROL_BUSY) {
					ast_log(DEBUG_MODE, "Found qualified busy signal\n");

					/* By default, we terminate on busy */
					if (!ignorebusydetect) {
						res = transfer_exten(chan, "BUSY", "busy");
						ast_frfree(fr);
						break;
					}
					gotbusy++;
				} else if (fr2->subclass == AST_CONTROL_ANSWER) {
					ast_log(DEBUG_MODE, "Found Asterisk's answer, but ignoring it\n");
					/* Forget about Asterisk's answer control for now */
				} else if (fr2->subclass == AST_CONTROL_RINGING) {
					ast_log(DEBUG_MODE, "Found qualified ring signal\n");

					/* If they are looking to detect only ring tone */
					if (returnonring) {
						res = transfer_exten(chan, "RING", "ring");
						ast_frfree(fr);
						break;
					}
					gotring++;
				} else if (fr2->subclass == AST_CONTROL_CONGESTION) {
					ast_log(DEBUG_MODE, "Found qualified congestion signal\n");

					/* By default, we terminate on congestion */
					if (!ignorecongestiondetect) {
						res = transfer_exten(chan, "CONGESTION", "congestion");
						ast_frfree(fr);
						break;
					}
					gotcongestion++;
				} 
#ifdef DSP_PROGRESS_DIALTONE
				else if (fr2->subclass == AST_CONTROL_DIALTONE) {
					ast_log(DEBUG_MODE, "Found qualified dialtone\n");

					/* If they are looking to detect only dial tone */
					if (returnondialtone) {
						res = transfer_exten(chan, "DIALTONE", "dialtone");
						ast_frfree(fr);
						break;
					}
					gotdialtone++;
				}               
#endif
#ifdef DSP_PROGRESS_DEAD
				else if (fr2->subclass == AST_CONTROL_DEAD) {
					ast_log(DEBUG_MODE, "Found qualified dead\n");

					/* If they are looking to detect only dial tone */
					if (!ignoredeaddetect) {
						res = transfer_exten(chan, "DEAD", "dead");
						ast_frfree(fr);
						break;
					}
					gotdead++;
				}               
#endif
#ifdef DSP_PROGRESS_PICKUP
				else if (fr2->subclass == AST_CONTROL_PICKUP) {
					ast_log(DEBUG_MODE, "Found qualified pickup\n");

					/* By default, we detect pickup */
					if (!ignorepickupdetect) {
						res = transfer_exten(chan, "PICKUP", "pickup");
						ast_frfree(fr);
						break;
					}
					gotpickup++;
				}               
#endif	
			} 
			
			if ((fr->frametype == AST_FRAME_VOICE) && (fr->subclass == AST_FORMAT_SLINEAR)) {
#ifdef DSP_PROGRESS_PICKUP
				if (gotring && gotpickup) {
					int totalsilence = 0;
					long ms = 0;
					res = ast_dsp_silence(dsp, fr, &totalsilence);

					if (res && (totalsilence > sildur)) {
						/* We've been quiet a little while */
						if (notsilent) {
							/* We had heard some talking */
							gettimeofday(&end, NULL);
							ms = (end.tv_sec - start.tv_sec) * 1000;
							ms += (end.tv_usec - start.tv_usec) / 1000;
							ms -= sildur;
							if (ms < 0) {
								ms = 0;
							}
							if ((ms > mindur) && ((maxdur < 0) || (ms < maxdur))) {
								ast_log(DEBUG_MODE, "Found qualified token of %ld ms\n", ms);
								ast_log(DEBUG_MODE, "Found qualified answer\n");

								/* By default, we detect answer */
								if (!ignoreanswerdetect) {
									res = transfer_exten(chan, "ANSWER", "answer");
									ast_frfree(fr);
									break;
								}
								gotanswer++;
							} else {
								ast_log(DEBUG_MODE, "Found unqualified token of %ld ms\n", ms);
							}
							notsilent = 0;
						} 
					}                       
					if (!notsilent) {
						/* Heard some audio, mark the begining of the token */
						gettimeofday(&start, NULL);
						ast_log(DEBUG_MODE, "Start of voice token\n");
						notsilent = 1;
						gottalk++;
					}
				}
#endif
			}
			ast_frfree(fr);
		}
	} else {
		ast_log(LOG_WARNING, "Could not answer channel '%s'\n", chan->name);
	}

	if (res > -1) {
		if (origrformat && ast_set_read_format(chan, origrformat)) {
			ast_log(LOG_WARNING, "Failed to restore read format for %s to %s\n", 
				chan->name, ast_getformatname(origrformat));
		}
	}

	if (dsp) {
		ast_dsp_free(dsp);
	}
    
    LOCAL_USER_REMOVE(u);
    
    return res;
}

int unload_module(void)
{
    STANDARD_HANGUP_LOCALUSERS;
    return ast_unregister_application(app);
}

int load_module(void)
{
    return ast_register_application(app, nv_detectline_exec, synopsis, descrip);
}

char *description(void)
{
    return tdesc;
}

int usecount(void)
{
    int res;
    STANDARD_USECOUNT(res);
    return res;
}

char *key()
{
    return ASTERISK_GPL_KEY;
}
